#pragma once

//-----------------------------------------------------------------------
// Copyright (c) 2017 Snap Inc.
//-----------------------------------------------------------------------

#ifndef GLES20STD_TEXTURE // Necessary over and above #pragma once because this header can be injected at export by Studio, and also included by core from within core includes.
#define GLES20STD_TEXTURE

// OpenGL math for depth conversion from depth texture value to eye space and vice versa:
// Depth to local (write value for depth texture):
//     depth = (z * m22 + m32) / (z * m23);
//     depth = depth * 0.5 + 0.5;
// Depth to global (read value from depth texture to eye space):
//     depth = depth * 2.0 - 1.0;
//     z = m32 / (depth * m23 - m22);
// Projection matrix terms (column major):
//     m22 = -(f + n) / (f - n)
//     m32 = -2fn / (f - n)
//     m23 = -1
// We hardcoded m23, so be careful if camera projection matrix will be changed in future

highp float depthToGlobal(highp float depth, highp vec2 projectionMatrixTerms) {
    highp float m22 = projectionMatrixTerms.x;
    highp float m32 = projectionMatrixTerms.y;
    depth = depth * 2.0 - 1.0;
    return m32 / (-depth - m22);
}

highp float depthToLocal(highp float depth, highp vec2 projectionMatrixTerms) {
    highp float m22 = projectionMatrixTerms.x;
    highp float m32 = projectionMatrixTerms.y;
    depth = depth != 0.0 ? (-m22 - m32 / depth) : 0.0;
    return depth * 0.5 + 0.5;
}

#ifdef GL_ES
#if __VERSION__ >= 300
#define SC_SUPPORT_2D_ARRAY_TEXTURE 1
#define SC_SUPPORT_3D_TEXTURE 1
#define texture2DArray texture
#define texture3D texture
#define textureCube texture
#else // #if __VERSION__ >= 300
#define SC_SUPPORT_2D_ARRAY_TEXTURE 0
#define SC_SUPPORT_3D_TEXTURE 0
#endif // #else // #if __VERSION__ >= 300
#else  // #ifdef GL_ES
#define SC_SUPPORT_2D_ARRAY_TEXTURE 1
#define SC_SUPPORT_3D_TEXTURE 1
#endif // #else #ifdef GL_ES

#define CLAMP_UV(uv, uvMin, uvMax, useClampToBorder, borderClampFactor) {\
    float clampedUv = clamp(uv, uvMin, uvMax);\
\
    if (useClampToBorder) {\
        borderClampFactor *= step(abs(uv - clampedUv), 0.00001);\
    }\
\
    uv = clampedUv;\
}

#define SOFTWARE_WRAP(uv, softwareWrapMode, useClampToBorder, clampToBorderFactor) {\
    if (softwareWrapMode == SOFTWARE_WRAP_MODE_CLAMP_TO_EDGE || softwareWrapMode == SOFTWARE_WRAP_MODE_CLAMP_TO_BORDER) {\
        CLAMP_UV(uv, 0.0, 1.0, useClampToBorder, clampToBorderFactor); \
    } else if (softwareWrapMode == SOFTWARE_WRAP_MODE_REPEAT) {\
        uv = fract(uv); \
    } else if (softwareWrapMode == SOFTWARE_WRAP_MODE_MIRRORED_REPEAT) {\
        float uvFract = fract(uv);\
        float uvInt = uv - uvFract;\
        float uvOdd = step(0.25, fract(uvInt * 0.5));\
        uv = mix(uvFract, 1.0 - uvFract, clamp(uvOdd, 0.0, 1.0));\
    }\
}

// Mediump devices have low UV precision, which can create strange artifacts like streaks at the edges of animated
// textures (see [LOOK-39049]). Since most people don't develop on mediump devices, we'd like to emulate these bugs
// in Studio, so that creators can fix the bugs during development.
#ifdef GL_ES
#define QUANTIZE_UV(A) A
#else
#define QUANTIZE_UV(A) A - (fract(A * 4096.0) / 4096.0);
#endif

#ifdef FRAGMENT_SHADER
#define TEXTURE2D(sampler, uv, mipBias) texture2D(sampler, uv, mipBias)
#define TEXTURE2DARRAY(sampler, uv, mipBias) texture2DArray(sampler, uv, mipBias)
#define TEXTURE3D(sampler, uv, mipBias) texture3D(sampler, uv, mipBias)
#define TEXTURECUBE(sampler, uv, mipBias) textureCube(sampler, uv, mipBias)
#else // #ifdef FRAGMENT_SHADER
#define TEXTURE2D(sampler, uv, mipBias) texture2D(sampler, uv)
#define TEXTURE2DARRAY(sampler, uv, mipBias) texture2DArray(sampler, uv)
#define TEXTURE3D(sampler, uv, mipBias) texture3D(sampler, uv)
#define TEXTURECUBE(sampler, uv, mipBias) textureCube(sampler, uv)
#endif // #else // #ifdef FRAGMENT_SHADER

// GPU_BUG_004 [LOOK-37728] 03/05/2019 GT-I9500 SGX 544 Android 4.4.2: The compiler and/or preprocessor on this device
// is broken, so it can’t handle the variable name “uv” being passed as an argument into a macro, which within its own
// scope also declares uv (MACRO(uvIn) {vec2 uv = uvIn;...}). The compiler complains that “uv” is used without being
// initialized. The workaround is to rename either one of the variables. We chose to rename the one inside the macro
// to uvTemp. This preprocessor is also broken in that it complains about our hash-macro and hash-hash concatenation
// constructs, even though they only appear inside a false ifdef branch. Other preprocessors ignore that stuff, this
// one chokes. This is not a real issue, since it only appears in the fallback case where the shader failed to compile
// and we attempt to compile without our preprocessor, and the compilation would fail anyway with or without our
// preprocessor (the fallback is not necessary anymore and will be removed in the near future).

#define SAMPLE_TEXTURE(tex, sampler, uvIn, useUvTransform, uvTransform, softwareWrapModeU, softwareWrapModeV, useUvMinMax, uvMinMax, useClampToBorder, borderColor, mipBias) {\
    vec2 uvTemp = uvIn;\
\
    float clampToBorderFactor = 1.0;\
\
    SOFTWARE_WRAP(uvTemp.x, softwareWrapModeU, useClampToBorder && !useUvMinMax, clampToBorderFactor);\
    SOFTWARE_WRAP(uvTemp.y, softwareWrapModeV, useClampToBorder && !useUvMinMax, clampToBorderFactor);\
\
    if (useUvMinMax) {\
        CLAMP_UV(uvTemp.x, uvMinMax.x, uvMinMax.z, useClampToBorder, clampToBorderFactor);\
        CLAMP_UV(uvTemp.y, uvMinMax.y, uvMinMax.w, useClampToBorder, clampToBorderFactor);\
    }\
\
    if (useUvTransform) {\
        uvTemp = vec2(uvTransform * vec3(uvTemp, 1.0));\
    }\
\
    uvTemp = QUANTIZE_UV(uvTemp); \
\
    tex = TEXTURE2D(sampler, uvTemp, mipBias);\
\
    if (useClampToBorder) {\
        tex = mix(borderColor, tex, clampToBorderFactor);\
    }\
}

#if SC_SUPPORT_2D_ARRAY_TEXTURE
#define SAMPLE_TEXTURE_2D_ARRAY(tex, sampler, uvwIn, useUvTransform, uvTransform, mipBias) {\
    vec2 uvTemp = uvwIn.xy;\
\
    if (useUvTransform) {\
        uvTemp = vec2(uvTransform * vec3(uvTemp, 1.0));\
    }\
\
    uvTemp = QUANTIZE_UV(uvTemp); \
\
    tex = TEXTURE2DARRAY(sampler, vec3(uvTemp, uvwIn.z), mipBias);\
}
#else
#define SAMPLE_TEXTURE_2D_ARRAY(tex, sampler, uvwIn, useUvTransform, uvTransform, mipBias) {\
    tex = vec4(0.0, 0.0, 0.0, 1.0);\
}
#endif

#if SC_SUPPORT_3D_TEXTURE
#define SAMPLE_TEXTURE_3D(tex, sampler, uvwIn, mipBias) {\
    vec3 uvwTemp = uvwIn;\
\
    uvwTemp = QUANTIZE_UV(uvwTemp); \
\
    tex = TEXTURE3D(sampler, uvwTemp, mipBias);\
}
#else
#define SAMPLE_TEXTURE_3D(tex, sampler, uvwIn, mipBias) {\
    tex = vec4(0.0, 0.0, 0.0, 1.0);\
}
#endif

#define SAMPLE_TEXTURE_CUBE(tex, sampler, uvwIn, mipBias) {\
    vec3 uvwTemp = uvwIn;\
\
    uvwTemp = QUANTIZE_UV(uvwTemp); \
\
    tex = TEXTURECUBE(sampler, uvwTemp, mipBias);\
}

#define SAMPLE_DEPTH_TEXTURE(tex, sampler, uvIn, useUvTransform, uvTransform) {\
    vec2 uvTemp = uvIn;\
\
    if (useUvTransform) {\
        uvTemp = vec2(uvTransform * vec3(uvTemp, 1.0));\
    }\
\
    uvTemp = QUANTIZE_UV(uvTemp); \
\
    tex = texture2D(sampler, uvTemp).r;\
}

#define SAMPLE_DEPTH_TEXTURE_NDC(tex, sampler, uvIn, useUvTransform, uvTransform, projectionMatrixTerms) {\
    vec2 uvTemp = uvIn;\
\
    if (useUvTransform) {\
        uvTemp = vec2(uvTransform * vec3(uvTemp, 1.0));\
    }\
\
    uvTemp = QUANTIZE_UV(uvTemp); \
\
    float depth = texture2D(sampler, uvTemp).r;\
\
    tex = depthToGlobal(depth, projectionMatrixTerms);\
}

#ifdef CUSTOM_PREPROCESSOR // This is never defined. It guards the glsl compiler from seeing invalid macro syntax. It's ignored by the custom preprocessor, which processes these macros as if the #ifdef wasn't here.

#macro DECLARE_TEXTURE(NAME)
uniform sampler2D NAME;
uniform mat3 NAME##Transform;
uniform vec4 NAME##UvMinMax;
uniform vec4 NAME##BorderColor;
uniform vec4 NAME##Size; // x, y correspond to width and height, z, w corresponding to 1/width, 1/height

// These defines are necessary because, in Studio, Core doesn't know what textures are available before the shader is compiled once, so it can't set these per-texture defines on the first comnpile, and the lack of these defines would cause compile errors.
#ifndef SC_USE_UV_TRANSFORM_##NAME
#define SC_USE_UV_TRANSFORM_##NAME false
#define SC_SOFTWARE_WRAP_MODE_U_##NAME SOFTWARE_WRAP_MODE_DISABLED
#define SC_SOFTWARE_WRAP_MODE_V_##NAME SOFTWARE_WRAP_MODE_DISABLED
#define SC_USE_UV_MIN_MAX_##NAME false
#endif
#ifndef SC_USE_CLAMP_TO_BORDER_##NAME
#define SC_USE_CLAMP_TO_BORDER_##NAME false
#endif

// Compatibility for code materials that don't set this define.
#ifndef NAME##UV
#define NAME##UV 0
#endif
#endmacro

#macro SAMPLE_TEX(NAME, UV, MIP_BIAS)
    vec4 NAME##Sample;
    SAMPLE_TEXTURE(NAME##Sample, NAME, UV, SC_USE_UV_TRANSFORM_##NAME, NAME##Transform, SC_SOFTWARE_WRAP_MODE_U_##NAME, SC_SOFTWARE_WRAP_MODE_V_##NAME, SC_USE_UV_MIN_MAX_##NAME, NAME##UvMinMax, SC_USE_CLAMP_TO_BORDER_##NAME, NAME##BorderColor, MIP_BIAS);
#endmacro

#macro DECLARE_TEXTURE_2D_ARRAY(NAME)
#if SC_SUPPORT_2D_ARRAY_TEXTURE

#if __VERSION__ >= 300
uniform sc_FragmentPrecision sampler2DArray NAME;
#else
uniform sampler2DArray NAME;
#endif
#else // #if SC_SUPPORT_2D_ARRAY_TEXTURE
uniform sampler2D NAME;
#endif // #if SC_SUPPORT_2D_ARRAY_TEXTURE

uniform mat3 NAME##Transform;
uniform vec4 NAME##Size; // x, y corresponding to width and height, z, w corresponding to 1/width, 1/height
uniform vec4 NAME##ArrayCount; // x corresponding to array count, y corresponding to (1 / array count)

// These defines are necessary because, in Studio, Core doesn't know what textures are available before the shader is compiled once, so it can't set these per-texture defines on the first comnpile, and the lack of these defines would cause compile errors.
#ifndef SC_USE_UV_TRANSFORM_##NAME
#define SC_USE_UV_TRANSFORM_##NAME false
#endif

// Compatibility for code materials that don't set this define.
#ifndef NAME##UV
#define NAME##UV 0
#endif

#endmacro

#macro SAMPLE_TEX_2D_ARRAY(NAME, UVW, MIP_BIAS)
    vec4 NAME##Sample;
    SAMPLE_TEXTURE_2D_ARRAY(NAME##Sample, NAME, UVW, SC_USE_UV_TRANSFORM_##NAME, NAME##Transform, MIP_BIAS);
#endmacro


#macro DECLARE_TEXTURE_3D(NAME)

#if SC_SUPPORT_3D_TEXTURE

#if (__VERSION__ >= 300)
uniform sc_FragmentPrecision sampler3D NAME;
#else
uniform sampler3D NAME;
#endif
#else // #if SC_SUPPORT_3D_TEXTURE
uniform sampler2D NAME;
#endif // #if SC_SUPPORT_3D_TEXTURE

uniform vec4 NAME##Size; // x, y coressponding to width and height, z, w corresponding to 1/width, 1/height
uniform vec2 NAME##Depth; // x coressponding to depth, y coressponding to 1/depth

// Compatibility for code materials that don't set this define.
#ifndef NAME##UV
#define NAME##UV 0
#endif

#endmacro

#macro SAMPLE_TEX_3D(NAME, UVW, MIP_BIAS)
    vec4 NAME##Sample;
    SAMPLE_TEXTURE_3D(NAME##Sample, NAME, UVW, MIP_BIAS);
#endmacro

#macro DECLARE_TEXTURE_CUBE(NAME)

#if (__VERSION__ >= 300)
uniform sc_FragmentPrecision samplerCube NAME;
#else
uniform samplerCube NAME;
#endif
uniform vec4 NAME##Size; // x, y correspond to width and height, z, w corresponding to 1/width, 1/height

// Compatibility for code materials that don't set this define.
#ifndef NAME##UV
#define NAME##UV 0
#endif

#endmacro

#macro SAMPLE_TEX_CUBE(NAME, UVW, MIP_BIAS)
vec4 NAME##Sample;
SAMPLE_TEXTURE_CUBE(NAME##Sample, NAME, UVW, MIP_BIAS);
#endmacro

#macro DECLARE_DEPTH_TEXTURE(NAME)
uniform sampler2D NAME;
uniform mat3 NAME##Transform;
uniform vec3 NAME##ProjectionMatrixTerms; // x - m22, y = m32
uniform vec4 NAME##UvMinMax;
uniform vec4 NAME##Size; // x, y correspond to width and height, z, w corresponding to 1/width, 1/height

#ifndef SC_USE_UV_TRANSFORM_##NAME
#define SC_USE_UV_TRANSFORM_##NAME false
#endif
#endmacro // DECLARE_DEPTH_TEXTURE

#macro SAMPLE_DEPTH_TEX(NAME, UV)
float NAME##Sample;
SAMPLE_DEPTH_TEXTURE(NAME##Sample, NAME, UV, SC_USE_UV_TRANSFORM_##NAME, NAME##Transform);
#endmacro // SAMPLE_DEPTH_TEX

#macro SAMPLE_DEPTH_TEX_NDC(NAME, UV)
float NAME##Sample;
SAMPLE_DEPTH_TEXTURE_NDC(NAME##Sample, NAME, UV, SC_USE_UV_TRANSFORM_##NAME, NAME##Transform, NAME##ProjectionMatrixTerms);
#endmacro // SAMPLE_DEPTH_TEX_NDC

#endif // #ifdef CUSTOM_PREPROCESSOR

#ifdef SC_TEXTURE_EXTERNAL
#define sc_SamplerExternalOES samplerExternalOES
#else
#define sc_SamplerExternalOES sampler2D
#endif // SC_TEXTURE_EXTERNAL

#ifdef FRAGMENT_SHADER

#if defined(GL_ES) && (__VERSION__ < 300) && !defined(GL_OES_standard_derivatives)
#define dFdx(A) (A)
#define dFdy(A) (A)
#define fwidth(A) (A)
#endif

#ifdef SC_TEXTURE_EXTERNAL
#define sc_textureExternalOESLod(tex, size, uv, lod) texture2D(tex, uv)
// Due to LOOK-66275, we cannot apply any lod with samplerExternalOES
#else
vec4 sc_texture2DLod(sampler2D tex, vec2 textureSize, vec2 uv, float lod) {
#if (__VERSION__ == 120)
#define SC_TESTURE2DLOD_IS_SUPPORTED
    return texture2DLod(tex, uv, lod);
#elif (__VERSION__ >= 300)
#define SC_TESTURE2DLOD_IS_SUPPORTED
    return textureLod(tex, uv, lod);
#elif defined(GL_EXT_shader_texture_lod)
#define SC_TESTURE2DLOD_IS_SUPPORTED
    return texture2DLodEXT(tex, uv, lod);
#elif defined(GL_OES_standard_derivatives)
#define SC_TESTURE2DLOD_IS_SUPPORTED
    vec2 texels = uv * textureSize;
    float dudx = dFdx(texels.x);
    float dvdx = dFdx(texels.y);
    float dudy = dFdy(texels.x);
    float dvdy = dFdy(texels.y);
    float rho = max(length(vec2(dudx, dvdx)), length(vec2(dudy, dvdy))); // OpenGL reference calculation
    float mu = max(abs(dudx), abs(dudy));
    float mv = max(abs(dvdx), abs(dvdy));
    float rho2 = max(mu, mv); // The allowed alternative OpoenGL reference calculation that seems to match the main reference best.
    float mip = log2(rho2);
    float bias = lod - mip;
    return texture2D(tex, uv, bias);
#else
    return texture2D(tex, uv, lod);
#endif
}

#define sc_textureExternalOESLod(tex, size, uv, lod) sc_texture2DLod(tex, size, uv, lod)
#endif // SC_TEXTURE_EXTERNAL

#if SC_SUPPORT_2D_ARRAY_TEXTURE
vec4 sc_texture2DArrayLod(sc_FragmentPrecision sampler2DArray tex, vec3 uv, float lod) {
#if (__VERSION__ == 120)
    return texture2DArrayLod(tex, uv, lod);
#elif (__VERSION__ >= 300)
    return textureLod(tex, uv, lod);
#elif defined(GL_EXT_shader_texture_lod)
    return texture2DArrayLodEXT(tex, uv, lod);
#else
    return texture2DArray(tex, uv, lod);
#endif
}
#else

#define sc_texture2DArrayLod(tex, uv, lod) vec4(0, 0, 0, 1)

#endif // SC_SUPPORT_2D_ARRAY_TEXTURE

#endif // FRAGMENT_SHADER

#if SC_SUPPORT_3D_TEXTURE
vec4 sc_texture3DLod(sc_FragmentPrecision sampler3D tex, vec3 uv, float lod) {
#if (__VERSION__ == 120)
    return texture3DLod(tex, uv, lod);
#elif (__VERSION__ >= 300)
    return textureLod(tex, uv, lod);
#elif defined(GL_EXT_shader_texture_lod)
    return texture3DLodEXT(tex, uv, lod);
#else
    return texture3D(tex, uv, lod);
#endif
}
#else

#define sc_texture3DLod(tex, uv, lod) vec4(0, 0, 0, 1)

#endif // SC_SUPPORT_3D_TEXTURE

vec4 sc_textureCubeLod(samplerCube tex, vec3 uv, float lod) {
#if (__VERSION__ < 120)
    return textureCube(tex, uv);
#elif (__VERSION__ == 120)
    return textureCubeLod(tex, uv, lod);
#elif (__VERSION__ >= 300)
    return textureLod(tex, uv, lod);
#elif defined(GL_EXT_shader_texture_lod)
    return textureCubeLodEXT(tex, uv, lod);
#else
    return textureCube(tex, uv, lod);
#endif
}

#endif // GLES20STD_TEXTURE
