//-----------------------------------------------------------------------
// Copyright (c) 2018 Snap Inc.
//-----------------------------------------------------------------------

#include <metal_stdlib>
using namespace metal;

constant bool MASK [[function_constant(0)]];
constant bool FXAA [[function_constant(1)]];
constant int MASK_CHANNEL[[function_constant(2)]];

struct Vertex
{
    float3 position [[attribute(0)]];
    float2 texture0 [[attribute(1)]];
};

struct ColorInOut {
    float4 position [[position]];
    half2 varTex0;
    half2 varTex1[[function_constant(FXAA)]];
    half2 varTex2[[function_constant(FXAA)]];
    half2 varTex3[[function_constant(FXAA)]];
    half2 varTex4[[function_constant(FXAA)]];
    half2 varTexBack[[function_constant(MASK)]];
    half2 varTexMask[[function_constant(MASK)]];
};

struct Uniforms
{
    float2 inputTextureScale;
    float3x3 inputTextureTransform;
    float3x3 maskTransform;
    float3x3 backTransform;
};

struct FSUniforms
{
    float4 backColorMult;
    float2 inputTextureScale;
};

// vertex shader function
vertex ColorInOut vs_main(Vertex vtx [[ stage_in ]],
                          constant Uniforms& uniforms [[ buffer(2) ]])
{
    ColorInOut out;
    out.position = float4(vtx.position, 1.0);
    
    float2 baseTex = (uniforms.inputTextureTransform * float3(vtx.texture0, 1.0)).xy;
    baseTex.y = 1.0-baseTex.y;// metal 
    out.varTex0 = half2(baseTex);
    if(FXAA)
    {
        out.varTex1 = half2(baseTex + (float2(-1.0, -1.0) * uniforms.inputTextureScale));
        out.varTex2 = half2(baseTex + (float2(+1.0, -1.0) * uniforms.inputTextureScale));
        out.varTex3 = half2(baseTex + (float2(-1.0, +1.0) * uniforms.inputTextureScale));
        out.varTex4 = half2(baseTex + (float2(+1.0, +1.0) * uniforms.inputTextureScale));
    }
    if( MASK)
    {
        out.varTexMask = half2((uniforms.maskTransform * float3(vtx.texture0, 1.0)).xy);
        out.varTexBack = half2((uniforms.backTransform * float3(vtx.texture0, 1.0)).xy);
    }
    return out;
}

// fragment shader function
fragment half4 fs_main(ColorInOut in [[stage_in]],
                                constant FSUniforms& uniforms,
                                texture2d<half> inputTexture,
                                sampler inputTextureSampler,
                                texture2d<half> maskTexture [[function_constant(MASK)]],
                                sampler maskTextureSampler [[function_constant(MASK)]],
                                texture2d<half> backTexture [[function_constant(MASK)]],
                                sampler backTextureSampler [[function_constant(MASK)]])
{
    half4 result;
    
    if(FXAA)
    {
        // The parameters are hardcoded for now, but could be
        // made into uniforms to control from the program.
        half FXAA_SPAN_MAX = 8.0h;
        half FXAA_REDUCE_MUL = 1.0h/8.0h;
        half FXAA_REDUCE_MIN = 1.0h/128.0h;
        
        half3 rgbNW = half3(inputTexture.sample(inputTextureSampler, float2(in.varTex1)).rgb);
        half3 rgbNE = half3(inputTexture.sample(inputTextureSampler, float2(in.varTex2)).rgb);
        half3 rgbSW = half3(inputTexture.sample(inputTextureSampler, float2(in.varTex3)).rgb);
        half3 rgbSE = half3(inputTexture.sample(inputTextureSampler, float2(in.varTex4)).rgb);
        half3 rgbM  = half3(inputTexture.sample(inputTextureSampler, float2(in.varTex0)).rgb);
        
        half3 luma = half3(0.299h, 0.587h, 0.114h);
        half lumaNW = dot(rgbNW, luma);
        half lumaNE = dot(rgbNE, luma);
        half lumaSW = dot(rgbSW, luma);
        half lumaSE = dot(rgbSE, luma);
        half lumaM  = dot( rgbM, luma);
        
        half lumaMin = min(lumaM, min(min(lumaNW, lumaNE), min(lumaSW, lumaSE)));
        half lumaMax = max(lumaM, max(max(lumaNW, lumaNE), max(lumaSW, lumaSE)));
        
        half2 dir;
        dir.x = -((lumaNW + lumaNE) - (lumaSW + lumaSE));
        dir.y =  ((lumaNW + lumaSW) - (lumaNE + lumaSE));
        
        half dirReduce = max((lumaNW + lumaNE + lumaSW + lumaSE) * (0.25h * FXAA_REDUCE_MUL), FXAA_REDUCE_MIN);
        
        half rcpDirMin = 1.0h/(min(abs(dir.x), abs(dir.y)) + dirReduce);
        
        dir = min(half2(FXAA_SPAN_MAX,  FXAA_SPAN_MAX),
                  max(half2(-FXAA_SPAN_MAX, -FXAA_SPAN_MAX), dir * rcpDirMin)) * half2(uniforms.inputTextureScale);
        
        half4 colorA = 0.5h * ( inputTexture.sample(inputTextureSampler, float2(in.varTex0 - dir * 1.0h/6.0h)) +
                             inputTexture.sample(inputTextureSampler, float2(in.varTex0 + dir * 1.0h/6.0h)));
        half4 colorB = colorA * 0.5h + 0.25h * (inputTexture.sample(inputTextureSampler, float2(in.varTex0 - dir * 0.5h)) +
                                             inputTexture.sample(inputTextureSampler, float2(in.varTex0 + dir * 0.5h)));
        half lumaB = dot(colorB.rgb, luma);
        
        if((lumaB < lumaMin) || (lumaB > lumaMax)) {
            result = colorA;
        } else {
            result = colorB;
        }
    }
    else
    {
        result = inputTexture.sample(inputTextureSampler, float2(in.varTex0));
    }

    if(MASK)
    {
        half4 mask;
        if( MASK_CHANNEL==0 )
        {
            mask = maskTexture.sample(maskTextureSampler, float2(in.varTexMask)).rrrr;
        }
        else if (MASK_CHANNEL==1 )
        {
            mask = maskTexture.sample(maskTextureSampler, float2(in.varTexMask));
        } else
        {
            mask = maskTexture.sample(maskTextureSampler, float2(in.varTexMask)).aaaa;
        }
        half4 back = backTexture.sample(backTextureSampler, float2(in.varTexBack))*(half4)uniforms.backColorMult;
        result = mix(back,result,mask);
    }
    return result;
};

