//-----------------------------------------------------------------------
// Copyright (c) 2018 Snap Inc.
//-----------------------------------------------------------------------

#include <metal_stdlib>
using namespace metal;

#if defined(CMDLINE)
#include "metal_std.metal"
#else
#include <metal_std.metal>
#endif

constant bool DEBUG_RATIO  = false;
constant bool DEBUG_COVERAGE = false;

// metal does not support dynamic array size controlled by constant, please keep this values equal to MtlGraphics constants
#define MaxLiquifyPointsHardwareLimit 10

constant float POWEREPSILON = 1.0/1280.0;

constant int MAX_LIQUIFY [[function_constant(FUNCONST_FS_BIAS+0)]];
constant bool CAMERA_ORTHO [[function_constant(FUNCONST_FS_BIAS+1)]];

struct Vertex
{
    float2 atbCoord2d [[attribute(0)]];
};

struct ColorInOut {
    float4 position [[position]];
    float2 varScreenPos;
    float2 varScreenSpacePointsPos0;
    float2 varScreenSpacePointsPos1;
    float2 varScreenSpacePointsPos2;
    float2 varScreenSpacePointsPos3;
    float2 varScreenSpacePointsPos4;
    float2 varScreenSpacePointsPos5;
    float2 varScreenSpacePointsPos6;
    float2 varScreenSpacePointsPos7;
    float2 varScreenSpacePointsPos8;
    float2 varScreenSpacePointsPos9;
};

struct Uniforms
{
    sc_Camera_t sc_Camera;
    float4x4 sc_ViewProjectionMatrixInverse;
    float4x4 ptsInvMat[MaxLiquifyPointsHardwareLimit];      // points modelInv matrix
    float3 camDirO[MaxLiquifyPointsHardwareLimit];            // camera dir in point object space
    float4 coeffs[MaxLiquifyPointsHardwareLimit];                  // ptx,pty, coef,rad
};

float2 calculateObjSpace(int i,float4 worldPos,constant Uniforms& uniforms)
{
    // compute vertex in object space
    float4 vertexPosObjSpace = uniforms.ptsInvMat[i] * worldPos;
    float3 viewDirObjSpace;
    float4 camPosObjSpace;
    if( CAMERA_ORTHO ) {
        viewDirObjSpace = uniforms.camDirO[i];
        camPosObjSpace = vertexPosObjSpace-float4(viewDirObjSpace,0);
    } else {
        camPosObjSpace = uniforms.ptsInvMat[i] * float4(uniforms.sc_Camera.position.xyz, 1.0);
        viewDirObjSpace = normalize(vertexPosObjSpace.xyz - camPosObjSpace.xyz);
    }
    float denom = viewDirObjSpace.z;
    float t = 10000.0;
    if (denom != 0.0) // should never happened (clip from CPU code)
    {
        // float t = (center - ray.origin).dot(normal) / denom;
        t = -camPosObjSpace.z / denom;
    }
    if (t < 0.0)
    {
        t = 10000.0;
    }
    // intersection in point object space
    return camPosObjSpace.xy + t * viewDirObjSpace.xy;

}

// vertex shader function
vertex ColorInOut vs_main(Vertex vtx [[ stage_in ]],
                          constant Uniforms& uniforms [[ buffer(2) ]])
{
    ColorInOut out;
    out.position = float4(vtx.atbCoord2d.x, vtx.atbCoord2d.y, 0.0, 1.0);
    out.varScreenPos = vtx.atbCoord2d;

    float4 screenPos = float4(vtx.atbCoord2d.x, vtx.atbCoord2d.y, -1.0, 1.0);
    
    // compute vertex location in world space
    float4 worldPos = uniforms.sc_ViewProjectionMatrixInverse * screenPos;
    worldPos.xyz /= worldPos.w;
    worldPos.w = 1.0;

    // for cycle unrolled - no array allowed to pass from vertex to fragment shader
    switch(MAX_LIQUIFY)
    {
        case 10: out.varScreenSpacePointsPos9 = calculateObjSpace(9,worldPos,uniforms);
        case 9: out.varScreenSpacePointsPos8 = calculateObjSpace(8,worldPos,uniforms);
        case 8: out.varScreenSpacePointsPos7 = calculateObjSpace(7,worldPos,uniforms);
        case 7: out.varScreenSpacePointsPos6 = calculateObjSpace(6,worldPos,uniforms);
        case 6: out.varScreenSpacePointsPos5 = calculateObjSpace(5,worldPos,uniforms);
        case 5: out.varScreenSpacePointsPos4 = calculateObjSpace(4,worldPos,uniforms);
        case 4: out.varScreenSpacePointsPos3 = calculateObjSpace(3,worldPos,uniforms);
        case 3: out.varScreenSpacePointsPos2 = calculateObjSpace(2,worldPos,uniforms);
        case 2: out.varScreenSpacePointsPos1 = calculateObjSpace(1,worldPos,uniforms);
        case 1: out.varScreenSpacePointsPos0 = calculateObjSpace(0,worldPos,uniforms);
    }
    return out;
}

float liquifyVector(float curDistanceSquared, float radiusSquared, float coeficient) {
  return pow(clamp(curDistanceSquared / radiusSquared, POWEREPSILON, 1.0), coeficient);
}

void calcOffsetForPoint(thread float2& result, int i, float2 varScreenSpacePointsPos,float2 varScreenPos, constant Uniforms& uniforms)
{
    float2 fromPointCenterVector;
    float dist = dot(varScreenSpacePointsPos, varScreenSpacePointsPos);
    fromPointCenterVector = varScreenPos - uniforms.coeffs[i].xy;
    float vecCoef = liquifyVector(dist, uniforms.coeffs[i].w, uniforms.coeffs[i].z) - 1.0;
    vecCoef *= step(dist,uniforms.coeffs[i].w); // step(dist,coeffs[i].w) := coeffs[i].w >= dist
    result += (fromPointCenterVector * vecCoef);
}

// fragment shader function
fragment half4 fs_main(ColorInOut in [[stage_in]],
                       constant Uniforms& uniforms [[ buffer(2) ]],
                       texture2d<half> sc_ScreenTexture,
                       sampler sc_ScreenTextureSampler)
{
    //return half4(1.0, 0.0, 0.0, 1.0);
    float2 posToFetch = in.varScreenPos;

    // for cycle unrolled - no array allowed to pass from vertex to fragment shader
    switch(MAX_LIQUIFY)
    {
        case 10: calcOffsetForPoint(posToFetch,9,in.varScreenSpacePointsPos9,in.varScreenPos,uniforms);
        case 9: calcOffsetForPoint(posToFetch,8,in.varScreenSpacePointsPos8,in.varScreenPos,uniforms);
        case 8: calcOffsetForPoint(posToFetch,7,in.varScreenSpacePointsPos7,in.varScreenPos,uniforms);
        case 7: calcOffsetForPoint(posToFetch,6,in.varScreenSpacePointsPos6,in.varScreenPos,uniforms);
        case 6: calcOffsetForPoint(posToFetch,5,in.varScreenSpacePointsPos5,in.varScreenPos,uniforms);
        case 5: calcOffsetForPoint(posToFetch,4,in.varScreenSpacePointsPos4,in.varScreenPos,uniforms);
        case 4: calcOffsetForPoint(posToFetch,3,in.varScreenSpacePointsPos3,in.varScreenPos,uniforms);
        case 3: calcOffsetForPoint(posToFetch,2,in.varScreenSpacePointsPos2,in.varScreenPos,uniforms);
        case 2: calcOffsetForPoint(posToFetch,1,in.varScreenSpacePointsPos1,in.varScreenPos,uniforms);
        case 1: calcOffsetForPoint(posToFetch,0,in.varScreenSpacePointsPos0,in.varScreenPos,uniforms);
    }

    if(DEBUG_RATIO)
    {
        float offset = length(posToFetch-in.varScreenPos) * 10.0;
        return half4(offset, offset, offset, 1.0);
    }
    else if( DEBUG_COVERAGE )
    {
        return half4(1.0, 0.0, 0.0, 1.0);
    }
    return sc_ScreenTexture.sample(sc_ScreenTextureSampler, posToFetch * float2(0.5,-0.5) + float2(0.5,0.5));
}




