/*
 A vertex-shader-based particle system. Position and other random attributes are computed entirely in the vertex shader.
 Attribute data consists of a 'part' identifier (which corner of the particle is this) and a 'particleID', which
 serves as a random seed.
 */

precision highp float;

uniform highp float time;
uniform highp float lastParticleEmissionTime;
uniform highp float maxParticleID;
uniform mediump float aspectRatio;
uniform mediump float seedOffset;

// uniform definitions in SCXWeather.h

uniform float emissionBurst;
uniform float skipParticleProbability;

uniform float fadeInOverTime;
uniform float fadeOutOverTime;
uniform float particleLifetime;
uniform float particleLifetimeVariance;

uniform vec4 emissionRect;
uniform float angle;
uniform float angleVariance;
uniform float velocity;
uniform float velocityVariance;

uniform int frameCount;
uniform float frameDuration;

uniform float imageAspectRatio;
uniform float scale;
uniform float scaleVariance;

uniform float flutterAmplitude;
uniform float flutterMinPeriod;
uniform float flutterMaxPeriod;

uniform float rotationsPerSecond;
uniform float rotationsPerSecondVariance;
uniform float rotateIn3D;

uniform vec4 color;
uniform float alphaVariance;

const float twoPi = 6.2831853072;

attribute float vertexPart;
/*
 vertex part: what part of the particle is this vertex?
 0: upper left
 1: lower left
 2: lower right
 3: upper right
 */

attribute highp float particleID;
/*
 particle ID: 0..1 identifier for this particle. Determines a particle's emission offset in a generation.
 */

varying lowp vec2 faceCoord; // [0..1]
varying lowp float kAlpha;

// 'canonical' GLSL PRNG from https://thebookofshaders.com/10/
// for best results, seed values should be in the -1..1 range
float random(vec2 st) {
    highp float x = fract(sin(dot(st.xy,
                         vec2(12.9898,78.233)))*
                 43758.5453123);
    return (x - 0.5) * 2.0;
}
float random_0_1(vec2 st) {
    highp float x = fract(sin(dot(st.xy,
                                  vec2(12.9898,78.233)))*
                          43758.5453123);
    return x;
}

// these 'feature seeds' are added to the particle instance seed to compute
// values for random particle-specific features.
// the PRNG works best in the region of -1..1, so they should be small:
const mediump float LIFETIME_FEATURE_SEED = 0.01;
const mediump float SCALE_FEATURE_SEED = 0.02;
const mediump float INITIAL_X_FEATURE_SEED = 0.03;
const mediump float INITIAL_Y_FEATURE_SEED = 0.04;
const mediump float ANGLE_FEATURE_SEED = 0.05;
const mediump float VELOCITY_FEATURE_SEED = 0.06;
const mediump float ROTATION_FEATURE_SEED = 0.07;
const mediump float ALPHA_FEATURE_SEED = 0.08;
const mediump float FLUTTER_PERIOD_FEATURE_SEED = 0.09;
const mediump float FLUTTER_OFFSET_FEATURE_SEED = 0.10;
const mediump float ROTATION_TIME_OFFSET_FEATURE_SEED = 0.11;
const mediump float SKIP_PARTICLE_FEATURE_SEED = 0.12;

float randWithVariance(highp vec2 particleSeed, float featureSeed, float value, float variance) {
    return value + variance * random(particleSeed + featureSeed);
}

vec2 rotateVec(vec2 vec, float angle) {
    return vec2(vec.x * cos(angle) - vec.y * sin(angle), vec.x * sin(angle) + vec.y * cos(angle));
}

float adjustTextureYCoordForSpriteAnimation(float y, float timeSinceEmission) {
    float fc = float(frameCount);
    float frameIndex = floor(mod(timeSinceEmission / frameDuration, fc));
    return (y + frameIndex) / fc;
}

void main()
{
    float generationDuration = particleLifetime + particleLifetimeVariance;
    float emissionOffsetWithinGeneration = pow(particleID / maxParticleID, emissionBurst) * generationDuration;
    
    float timeSinceEmission = mod(time - emissionOffsetWithinGeneration, generationDuration);
    float emissionTime = time - timeSinceEmission;
    float generation = floor(emissionTime / generationDuration);
    highp vec2 particleSeed = vec2(particleID, 1.0 / (generation + 1.0)) + seedOffset;
    
    float lifetime = randWithVariance(particleSeed, LIFETIME_FEATURE_SEED, particleLifetime, particleLifetimeVariance);
    
    if (emissionTime < 0.0 || (lastParticleEmissionTime > 0.0 && emissionTime > lastParticleEmissionTime)) {
        kAlpha = 0.0;
    } else {
        kAlpha = randWithVariance(particleSeed, ALPHA_FEATURE_SEED, 1.0, alphaVariance);
        kAlpha *= clamp(timeSinceEmission / fadeInOverTime, 0.0, 1.0);
        kAlpha *= clamp((lifetime - timeSinceEmission) / fadeOutOverTime, 0.0, 1.0);
    }
    
    if (random_0_1(particleSeed + SKIP_PARTICLE_FEATURE_SEED) < skipParticleProbability) {
        kAlpha = 0.0;
    }
    
    mediump float scale = randWithVariance(particleSeed, SCALE_FEATURE_SEED, scale, scaleVariance);

    vec2 initialPosition = emissionRect.xy + vec2(random_0_1(particleSeed + INITIAL_X_FEATURE_SEED), random_0_1(particleSeed + INITIAL_Y_FEATURE_SEED)) * emissionRect.zw;
    mediump float particleAngle = randWithVariance(particleSeed, ANGLE_FEATURE_SEED, angle, angleVariance);
    mediump float particleVelocity = randWithVariance(particleSeed, VELOCITY_FEATURE_SEED, velocity, velocityVariance);
    mediump float distance = generationDuration * particleVelocity;
    
    vec2 displacement = vec2(0.0, distance);
    if (flutterAmplitude > 0.0) {
        float flutterPeriod = mix(flutterMinPeriod, flutterMaxPeriod, random_0_1(particleSeed + FLUTTER_PERIOD_FEATURE_SEED));
        float flutterPhase = random_0_1(particleSeed + FLUTTER_OFFSET_FEATURE_SEED) * twoPi;
        displacement.x += sin(timeSinceEmission / flutterPeriod + flutterPhase) * flutterAmplitude;
    }
    displacement = rotateVec(displacement, particleAngle);
    
    vec2 finalPosition = initialPosition + displacement;
    vec2 pos = mix(initialPosition, finalPosition, timeSinceEmission / generationDuration);
    
    mediump float particleRotationPerSecond = randWithVariance(particleSeed, ROTATION_FEATURE_SEED, rotationsPerSecond, rotationsPerSecondVariance);
    float rotationTimeOffset = timeSinceEmission + generationDuration * random_0_1(particleSeed + ROTATION_TIME_OFFSET_FEATURE_SEED);
    mediump float particleRotation = particleRotationPerSecond > 0.0 ? particleRotationPerSecond * rotationTimeOffset : 0.0;
    mediump float rotation3D = particleRotation * rotateIn3D;
    
    mediump float spriteRotation = particleRotation + particleAngle;
    
    lowp int part = int(vertexPart);
    float u = (part == 0 || part == 1) ? 0.0 : 1.0;
    float v = (part == 0 || part == 2) ? 0.0 : 1.0;

    vec2 localPos = vec2((u - 0.5) * scale * imageAspectRatio, (v - 0.5) * scale);
    // give the effect of 3d location without projection by scaling along one axis:
    localPos.x *= cos(rotation3D);
    // apply 2d rotation:
    localPos = rotateVec(localPos, spriteRotation);

    vec2 finalPos = pos + localPos;

    gl_Position = vec4(finalPos.x / aspectRatio, finalPos.y, 0.0, 1.0);
    float vAdjusted = adjustTextureYCoordForSpriteAnimation(v, timeSinceEmission);
    faceCoord = vec2(u, vAdjusted);
}
